import libcst
import random

from swesmith.constants import DEFAULT_PM_LIKELIHOOD, CodeEntity


class PythonProceduralModifier(libcst.CSTTransformer):
    min_complexity: int = 3
    max_complexity: int = float("inf")

    def __init__(self, likelihood: float = DEFAULT_PM_LIKELIHOOD, seed: float = 24):
        super().__init__()
        assert 0 <= likelihood <= 1, "Likelihood must be between 0 and 1."
        self.rand = random.Random(seed)
        self.likelihood = likelihood

    def flip(self) -> bool:
        return self.rand.random() < self.likelihood

    def can_change(self, code_entity: CodeEntity) -> bool:
        """Check if the CodeEntity satisfies the conditions of the modifier."""
        return (
            all(c in code_entity._tags for c in self.conditions)
            and self.min_complexity <= code_entity.complexity <= self.max_complexity
        )



import libcst

from swesmith.bug_gen.procedural import PythonProceduralModifier
from swesmith.constants import CodeProperty


class ControlIfElseInvertModifier(PythonProceduralModifier):
    explanation: str = (
        "The if-else conditions may be out of order, or the bodies are inverted."
    )
    name: str = "func_pm_ctrl_invert_if"
    conditions: list = [
        CodeProperty.IS_FUNCTION,
        CodeProperty.HAS_IF_ELSE,
    ]
    min_complexity: int = 5

    def leave_If(self, original_node: libcst.If, updated_node: libcst.If) -> libcst.If:
        if not self.flip():
            return updated_node

        # Only proceed if there's an else branch to swap with
        if not updated_node.orelse:
            return updated_node

        # We need to handle standard else blocks
        if isinstance(updated_node.orelse, libcst.Else):
            # Store the original bodies
            if_body = updated_node.body
            else_body = updated_node.orelse.body

            # Create a new else clause with the original if body
            new_else = libcst.Else(
                body=if_body,
                whitespace_before_colon=updated_node.orelse.whitespace_before_colon,
            )

            # Return a new If statement with swapped bodies
            return updated_node.with_changes(body=else_body, orelse=new_else)

        # Skip elif cases for now
        return updated_node


class ControlShuffleLinesModifier(PythonProceduralModifier):
    explanation: str = "The lines inside a function may be out of order."
    name: str = "func_pm_ctrl_shuffle"
    conditions: list = [
        CodeProperty.IS_FUNCTION,
        CodeProperty.HAS_LOOP,
    ]
    max_complexity: int = 10

    def leave_FunctionDef(
        self, original_node: libcst.FunctionDef, updated_node: libcst.FunctionDef
    ) -> libcst.FunctionDef:
        # Skip modification if random check fails
        if not self.flip():
            return updated_node

        # Make sure we're working with an indented block
        if not isinstance(updated_node.body, libcst.IndentedBlock):
            return updated_node

        # Get the body statements
        body = list(updated_node.body.body)

        # Don't shuffle if there are fewer than 2 statements
        if len(body) < 2:
            return updated_node

        # Create a shuffled copy of the statements
        shuffled_body = body.copy()
        self.rand.shuffle(shuffled_body)

        # Create a new indented block with the shuffled statements
        new_body = libcst.IndentedBlock(
            body=tuple(shuffled_body),
            indent=updated_node.body.indent,
            header=updated_node.body.header,
            footer=updated_node.body.footer,
        )

        # Return the updated function with the new body
        return updated_node.with_changes(body=new_body)



import re

from dataclasses import dataclass
from swesmith.constants import KEY_PATCH
from swebench.harness.constants import TestStatus
from swesmith.profiles.base import RepoProfile, registry
from swesmith.profiles.utils import X11_DEPS
from unidiff import PatchSet


@dataclass
class JavaScriptProfile(RepoProfile):
    """
    Profile for JavaScript repositories.
    """


def default_npm_install_dockerfile(mirror_name: str, node_version: str = "18") -> str:
    return f"""FROM node:{node_version}-bullseye
RUN apt update && apt install -y git  
RUN git clone https://github.com/{mirror_name} /testbed
WORKDIR /testbed
RUN npm install
"""


def parse_log_jest(log: str) -> dict[str, str]:
    """
    Parser for test logs generated with Jest. Assumes --verbose flag.

    Args:
        log (str): log content
    Returns:
        dict: test case to test status mapping
    """
    test_status_map = {}

    pattern = r"^\s*(✓|✕|○)\s(.+?)(?:\s\((\d+\s*m?s)\))?$"

    for line in log.split("\n"):
        match = re.match(pattern, line.strip())
        if match:
            status_symbol, test_name, _duration = match.groups()
            if status_symbol == "✓":
                test_status_map[test_name] = TestStatus.PASSED.value
            elif status_symbol == "✕":
                test_status_map[test_name] = TestStatus.FAILED.value
            elif status_symbol == "○":
                test_status_map[test_name] = TestStatus.SKIPPED.value
    return test_status_map


def parse_log_mocha(log: str) -> dict[str, str]:
    test_status_map = {}
    pattern = r"^\s*(✔|✖|-)\s(.+?)(?:\s\((\d+\s*m?s)\))?$"
    for line in log.split("\n"):
        match = re.match(pattern, line.strip())
        if match:
            status_symbol, test_name, _duration = match.groups()
            if status_symbol == "✔":
                test_status_map[test_name] = TestStatus.PASSED.value
            elif status_symbol == "✖":
                test_status_map[test_name] = TestStatus.FAILED.value
            elif status_symbol == "-":
                test_status_map[test_name] = TestStatus.SKIPPED.value
    return test_status_map


def parse_log_vitest(log: str) -> dict[str, str]:
    test_status_map = {}
    patterns = [
        (r"^✓\s+(.+?)(?:\s+\([\.\d]+ms\))?$", TestStatus.PASSED.value),
        (r"^✗\s+(.+?)(?:\s+\([\.\d]+ms\))?$", TestStatus.FAILED.value),
        (r"^○\s+(.+?)(?:\s+\([\.\d]+ms\))?$", TestStatus.SKIPPED.value),
        (r"^✓\s+(.+?)$", TestStatus.PASSED.value),
        (r"^✗\s+(.+?)$", TestStatus.FAILED.value),
        (r"^○\s+(.+?)$", TestStatus.SKIPPED.value),
    ]
    for line in log.split("\n"):
        for pattern, status in patterns:
            match = re.match(pattern, line.strip())
            if match:
                test_name = match.group(1).strip()
                test_status_map[test_name] = status
                break

    return test_status_map


@dataclass
class ReactPDFee5c96b8(JavaScriptProfile):
    owner: str = "diegomura"
    repo: str = "react-pdf"
    commit: str = "ee5c96b80326ba4441b71be4c7a85ba9f61d4174"
    test_cmd: str = "./node_modules/.bin/vitest --no-color --reporter verbose"

    @property
    def dockerfile(self):
        return f"""FROM node:20-bullseye
RUN apt update && apt install -y pkg-config build-essential libpixman-1-0 libpixman-1-dev libcairo2-dev libpango1.0-dev libjpeg-dev libgif-dev librsvg2-dev
RUN git clone https://github.com/{self.mirror_name} /testbed
WORKDIR /testbed
RUN yarn install
"""

    def log_parser(self, log: str) -> dict[str, str]:
        test_status_map = {}
        for line in log.split("\n"):
            for pattern, status in [
                (r"^\s*✓\s(.*)\s\d+ms", TestStatus.PASSED.value),
                (r"^\s*✗\s(.*)\s\d+ms", TestStatus.FAILED.value),
                (r"^\s*✖\s(.*)", TestStatus.FAILED.value),
                (r"^\s*✓\s(.*)", TestStatus.PASSED.value),
            ]:
                match = re.match(pattern, line)
                if match:
                    test_name = match.group(1).strip()
                    test_status_map[test_name] = status
                    break
        return test_status_map


@dataclass
class Markeddbf29d91(JavaScriptProfile):
    owner: str = "markedjs"
    repo: str = "marked"
    commit: str = "dbf29d9171a28da21f06122d643baf4e5d4266d4"
    test_cmd: str = "NO_COLOR=1 node --test"

    @property
    def dockerfile(self):
        return f"""FROM node:24-bullseye
RUN apt update && apt install -y git {X11_DEPS}
RUN git clone https://github.com/{self.mirror_name} /testbed
WORKDIR /testbed
RUN npm install
RUN npm test
"""

    def log_parser(self, log: str) -> dict[str, str]:
        test_status_map = {}
        fail_pattern = r"^\s*✖\s(.*?)\s\([\.\d]+ms\)"
        pass_pattern = r"^\s*✔\s(.*?)\s\([\.\d]+ms\)"
        for line in log.split("\n"):
            fail_match = re.match(fail_pattern, line)
            if fail_match:
                test = fail_match.group(1)
                test_status_map[test.strip()] = TestStatus.FAILED.value
            else:
                pass_match = re.match(pass_pattern, line)
                if pass_match:
                    test = pass_match.group(1)
                    test_status_map[test.strip()] = TestStatus.PASSED.value
        return test_status_map


@dataclass
class Babel2ea3fc8f(JavaScriptProfile):
    owner: str = "babel"
    repo: str = "babel"
    commit: str = "2ea3fc8f9b33a911840f17fbc407e7bfae2ed66f"
    test_cmd: str = "yarn jest --verbose"

    @property
    def dockerfile(self):
        return f"""FROM node:20-bullseye
RUN apt update && apt install -y git
RUN git clone https://github.com/{self.mirror_name} /testbed
WORKDIR /testbed
RUN make bootstrap
RUN make build
"""

    def log_parser(self, log: str) -> dict[str, str]:
        return parse_log_jest(log)

    def get_test_cmd(self, instance: dict, f2p_only: bool = False):
        if KEY_PATCH not in instance:
            return self.test_cmd, []
        test_folders = []
        for f in PatchSet(instance[KEY_PATCH]):
            parts = f.path.split("/")
            if len(parts) >= 2 and parts[0] == "packages":
                test_folders.append("/".join(parts[:2]))
        return f"{self.test_cmd} {' '.join(test_folders)}", test_folders


@dataclass
class GithubReadmeStats3e974011(JavaScriptProfile):
    owner: str = "anuraghazra"
    repo: str = "github-readme-stats"
    commit: str = "3e97401177143bb35abb42279a13991cbd584ca3"  # Using master branch, can be updated to specific commit hash
    test_cmd: str = "npm test -- --verbose"

    @property
    def dockerfile(self):
        return default_npm_install_dockerfile(self.mirror_name)

    def log_parser(self, log: str) -> dict[str, str]:
        return parse_log_jest(log)


@dataclass
class Mongoose5f57a5bb(JavaScriptProfile):
    owner: str = "Automattic"
    repo: str = "mongoose"
    commit: str = "5f57a5bbb2e8dfed8d04be47cdd17728633c44c1"  # Replace with a specific commit hash
    test_cmd: str = "npm test -- --verbose"

    @property
    def dockerfile(self):
        return default_npm_install_dockerfile(self.mirror_name)

    def log_parser(self, log: str) -> dict[str, str]:
        return parse_log_mocha(log)


@dataclass
class Axiosef36347f(JavaScriptProfile):
    owner: str = "axios"
    repo: str = "axios"
    commit: str = "ef36347fb559383b04c755b07f1a8d11897fab7f"  # Replace with a specific commit hash
    test_cmd: str = "npm run test:mocha -- --verbose"

    @property
    def dockerfile(self):
        return default_npm_install_dockerfile(self.mirror_name)

    def log_parser(self, log: str) -> dict[str, str]:
        return parse_log_mocha(log)


@dataclass
class Async23dbf76a(JavaScriptProfile):
    owner: str = "caolan"
    repo: str = "async"
    commit: str = (
        "23dbf76aeb04c7c3dd56276115b277e3fa9dd5cc"  # Replace with a real commit hash
    )
    test_cmd: str = "npm run mocha-node-test -- --verbose"

    @property
    def dockerfile(self):
        return default_npm_install_dockerfile(self.mirror_name)

    def log_parser(self, log: str) -> dict[str, str]:
        return parse_log_mocha(log)


@dataclass
class Expressef5f2e13(JavaScriptProfile):
    owner: str = "expressjs"
    repo: str = "express"
    commit: str = "ef5f2e13ef64a1575ce8c2d77b180d593644ccfa"
    test_cmd: str = "npm test -- --verbose"

    @property
    def dockerfile(self):
        return default_npm_install_dockerfile(self.mirror_name)

    def log_parser(self, log: str) -> dict[str, str]:
        return parse_log_mocha(log)


@dataclass
class Dayjsc8a26460(JavaScriptProfile):
    owner: str = "iamkun"
    repo: str = "dayjs"
    commit: str = "c8a26460d89a2ee9a7d3b9cafa124ea856ee883f"
    test_cmd: str = "npm test -- --verbose"

    @property
    def dockerfile(self):
        return default_npm_install_dockerfile(self.mirror_name)

    def log_parser(self, log: str) -> dict[str, str]:
        return parse_log_jest(log)


@dataclass
class Svelte6c9717a9(JavaScriptProfile):
    owner: str = "sveltejs"
    repo: str = "svelte"
    commit: str = "6c9717a91f2f6ae10641d1cf502ba13d227fbe45"
    test_cmd: str = "pnpm test -- --verbose"

    @property
    def dockerfile(self):
        return f"""FROM node:18-bullseye
RUN apt update && apt install -y git
RUN npm install -g pnpm@10.4.0
RUN git clone https://github.com/{self.mirror_name} /testbed
WORKDIR /testbed
RUN pnpm install
RUN pnpm playwright install chromium
RUN pnpm exec playwright install-deps
"""

    def log_parser(self, log: str) -> dict[str, str]:
        return parse_log_vitest(log)


@dataclass
class Commanderjs395cf714(JavaScriptProfile):
    owner: str = "tj"
    repo: str = "commander.js"
    commit: str = "395cf7145fe28122f5a69026b310e02df114f907"
    test_cmd: str = "npm test -- --verbose"

    @property
    def dockerfile(self):
        return default_npm_install_dockerfile(self.mirror_name, node_version="20")

    def log_parser(self, log: str) -> dict[str, str]:
        return parse_log_jest(log)


# Register all JavaScript profiles with the global registry
for name, obj in list(globals().items()):
    if (
        isinstance(obj, type)
        and issubclass(obj, JavaScriptProfile)
        and obj.__name__ != "JavaScriptProfile"
    ):
        registry.register_profile(obj)



"""
Purpose: Given a repository, procedurally generate a variety of bugs for functions/classes/objects in the repository.

Usage: python -m swesmith.bug_gen.procedural.generate \
    --repo <repo> \
    --commit <commit>
"""

import argparse
import json
import libcst
import random
import shutil

from pathlib import Path
from rich import print
from swesmith.bug_gen.utils import (
    apply_code_change,
    get_bug_directory,
    get_patch,
)
from swesmith.constants import (
    LOG_DIR_BUG_GEN,
    PREFIX_BUG,
    PREFIX_METADATA,
    BugRewrite,
    CodeEntity,
)
from swesmith.profiles import registry
from tqdm.auto import tqdm

from swesmith.bug_gen.procedural import PythonProceduralModifier
from swesmith.bug_gen.procedural.classes import (
    ClassRemoveBasesModifier,
    ClassRemoveFuncsModifier,
    ClassShuffleMethodsModifier,
)
from swesmith.bug_gen.procedural.control_flow import (
    ControlIfElseInvertModifier,
    ControlShuffleLinesModifier,
)
from swesmith.bug_gen.procedural.operations import (
    OperationBreakChainsModifier,
    OperationChangeConstantsModifier,
    OperationChangeModifier,
    OperationSwapOperandsModifier,
)
from swesmith.bug_gen.procedural.remove import (
    RemoveAssignModifier,
    RemoveConditionalModifier,
    RemoveLoopModifier,
    RemoveWrapperModifier,
)

PM_TECHNIQUES = [
    ClassRemoveBasesModifier(likelihood=0.25),
    ClassRemoveFuncsModifier(likelihood=0.15),
    ClassShuffleMethodsModifier(likelihood=0.25),
    ControlIfElseInvertModifier(likelihood=0.25),
    ControlShuffleLinesModifier(likelihood=0.25),
    RemoveAssignModifier(likelihood=0.25),
    RemoveConditionalModifier(likelihood=0.25),
    RemoveLoopModifier(likelihood=0.25),
    RemoveWrapperModifier(likelihood=0.25),
    OperationBreakChainsModifier(likelihood=0.4),
    OperationChangeConstantsModifier(likelihood=0.4),
    OperationChangeModifier(likelihood=0.4),
    OperationSwapOperandsModifier(likelihood=0.4),
]


def _process_candidate(
    candidate: CodeEntity, pm: PythonProceduralModifier, log_dir: Path, repo: str
):
    """
    Process a candidate by applying a given procedural modification to it.
    """
    # Apply transformation
    try:
        module = libcst.parse_module(candidate.src_code)
    except Exception:
        # Failed to parse code
        return False

    changed = False
    try:
        for _ in range(5):
            modified = module.visit(pm)
            if module.code != modified.code:
                changed = True
                break
    except Exception:
        return False

    if not changed:
        return False

    # Get modified function
    bug = BugRewrite(
        rewrite=modified.code,
        explanation=pm.explanation,
        cost=0.0,
        strategy=pm.name,
    )

    # Create artifacts
    bug_dir = get_bug_directory(log_dir, candidate)
    bug_dir.mkdir(parents=True, exist_ok=True)
    uuid_str = f"{pm.name}__{bug.get_hash()}"
    metadata_path = f"{PREFIX_METADATA}__{uuid_str}.json"
    bug_path = f"{PREFIX_BUG}__{uuid_str}.diff"

    with open(bug_dir / metadata_path, "w") as f:
        json.dump(bug.to_dict(), f, indent=2)
    apply_code_change(candidate, bug)
    patch = get_patch(repo, reset_changes=True)
    if patch:
        with open(bug_dir / bug_path, "w") as f:
            f.write(patch)
        return True
    return False


def main(
    repo: str,
    max_bugs: int,
    seed: int,
):
    random.seed(seed)
    total = 0
    rp = registry.get(repo)
    rp.clone()
    entities = rp.extract_entities()
    print(f"Found {len(entities)} entities in {repo}.")
    for pm in PM_TECHNIQUES:
        print(f"Generating [bold blue]{pm.name}[/bold blue] bugs in {repo}...")
        candidates = [x for x in entities if pm.can_change(x)]
        if not candidates:
            print(f"No candidates found in {repo}.")
            continue
        print(f"Found {len(candidates)} candidates in {repo}.")

        log_dir = LOG_DIR_BUG_GEN / repo
        log_dir.mkdir(parents=True, exist_ok=True)
        print(f"Logging bugs to {log_dir}")

        if max_bugs > 0 and len(candidates) > max_bugs:
            candidates = random.sample(candidates, max_bugs)

        for candidate in tqdm(candidates):
            total += _process_candidate(candidate, pm, log_dir, repo)

    shutil.rmtree(repo)
    print(f"Generated {total} bugs for {repo}.")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Generate bugs for a given repository and commit."
    )
    parser.add_argument(
        "repo",
        type=str,
        help="Name of a SWE-smith repository to generate bugs for.",
    )
    parser.add_argument(
        "--seed",
        type=int,
        default=24,
        help="Seed for random number generator.",
    )
    parser.add_argument(
        "--max_bugs",
        type=int,
        default=-1,
        help="Maximum number of bugs to generate.",
    )

    args = parser.parse_args()
    main(**vars(args))



import libcst

from swesmith.bug_gen.procedural import PythonProceduralModifier
from swesmith.constants import CodeProperty


class RemoveLoopModifier(PythonProceduralModifier):
    explanation: str = "There is one or more missing loops that is causing the bug."
    name: str = "func_pm_remove_loop"
    conditions: list = [
        CodeProperty.IS_FUNCTION,
        CodeProperty.HAS_LOOP,
    ]

    def leave_For(self, original_node, updated_node):
        return libcst.RemoveFromParent() if self.flip() else updated_node

    def leave_While(self, original_node, updated_node):
        return libcst.RemoveFromParent() if self.flip() else updated_node


class RemoveConditionalModifier(PythonProceduralModifier):
    explanation: str = "There is one or more missing conditionals that causes the bug."
    name: str = "func_pm_remove_cond"
    conditions: list = [
        CodeProperty.IS_FUNCTION,
        CodeProperty.HAS_IF,
    ]

    def leave_If(self, original_node, updated_node):
        return libcst.RemoveFromParent() if self.flip() else updated_node


class RemoveAssignModifier(PythonProceduralModifier):
    explanation: str = "There is likely a missing assignment in the code."
    name: str = "func_pm_remove_assign"
    conditions: list = [
        CodeProperty.IS_FUNCTION,
        CodeProperty.HAS_ASSIGNMENT,
    ]

    def leave_Assign(self, original_node, updated_node):
        return libcst.RemoveFromParent() if self.flip() else updated_node

    def leave_AugAssign(self, original_node, updated_node):
        return libcst.RemoveFromParent() if self.flip() else updated_node


class RemoveWrapperModifier(PythonProceduralModifier):
    explanation: str = "There are missing wrappers (with, try blocks) in the code."
    name: str = "func_pm_remove_wrapper"
    conditions: list = [
        CodeProperty.IS_FUNCTION,
        CodeProperty.HAS_WRAPPER,
    ]

    def leave_With(self, original_node, updated_node):
        return libcst.RemoveFromParent() if self.flip() else updated_node

    def leave_AsyncWith(self, original_node, updated_node):
        return libcst.RemoveFromParent() if self.flip() else updated_node

    def leave_Try(self, original_node, updated_node):
        return libcst.RemoveFromParent() if self.flip() else updated_node



from dataclasses import dataclass

from swebench.harness.constants import TestStatus
from swesmith.profiles.base import RepoProfile, registry


@dataclass
class RustProfile(RepoProfile):
    """
    Profile for Rust repositories.
    """

    test_cmd: str = "cargo test --verbose"

    def log_parser(self, log: str):
        test_status_map = {}
        for line in log.split("\n"):
            if "... ok" in line:
                test_name = line.rsplit(" ... ", 1)[0].strip()
                test_status_map[test_name] = TestStatus.PASSED.value
            elif "... FAILED" in line:
                test_name = line.rsplit(" ... ", 1)[0].strip()
                test_status_map[test_name] = TestStatus.FAILED.value
        return test_status_map


@dataclass
class Jsoncd55b5a0(RustProfile):
    owner: str = "serde-rs"
    repo: str = "json"
    commit: str = "cd55b5a0ff5f88f1aeb7a77c1befc9ddb3205201"

    @property
    def dockerfile(self):
        return f"""FROM rust:1.88
ARG DEBIAN_FRONTEND=noninteractive
ENV TZ=Etc/UTC

RUN apt update && apt install -y wget git build-essential \
&& rm -rf /var/lib/apt/lists/*

RUN git clone https://github.com/{self.mirror_name} /testbed
WORKDIR /testbed
RUN cargo test --verbose
"""


# Register all Rust profiles with the global registry
for name, obj in list(globals().items()):
    if (
        isinstance(obj, type)
        and issubclass(obj, RustProfile)
        and obj.__name__ != "RustProfile"
    ):
        registry.register_profile(obj)



import libcst

from swesmith.bug_gen.procedural import PythonProceduralModifier
from swesmith.constants import CodeProperty


class ClassRemoveBasesModifier(PythonProceduralModifier):
    explanation: str = "The base class has been removed from the class definition."
    name: str = "func_pm_class_rm_base"
    conditions: list = [CodeProperty.IS_CLASS, CodeProperty.HAS_PARENT]
    min_complexity: int = 10

    def leave_ClassDef(self, original_node, updated_node):
        bases = list(updated_node.bases)
        if len(bases) > 0 and self.flip():
            if len(bases) == 1:
                bases = []
            else:
                to_remove = self.rand.randint(0, len(bases) - 1)
                bases.pop(to_remove)

        return updated_node.with_changes(bases=tuple(bases))


class ClassShuffleMethodsModifier(PythonProceduralModifier):
    explanation: str = "The methods in a class have been shuffled."
    name: str = "func_pm_class_shuffle_funcs"
    conditions: list = [CodeProperty.IS_CLASS]
    min_complexity: int = 10

    def leave_ClassDef(self, original_node, updated_node):
        methods = [
            n for n in updated_node.body.body if isinstance(n, libcst.FunctionDef)
        ]
        non_methods = [
            n for n in updated_node.body.body if not isinstance(n, libcst.FunctionDef)
        ]
        self.rand.shuffle(methods)
        new_body = non_methods + methods
        return updated_node.with_changes(
            body=updated_node.body.with_changes(body=tuple(new_body))
        )


class ClassRemoveFuncsModifier(PythonProceduralModifier):
    explanation: str = (
        "Method(s) and their reference(s) have been removed from the class."
    )
    name: str = "func_pm_class_rm_funcs"
    conditions: list = [CodeProperty.IS_CLASS]
    min_complexity: int = 10

    def leave_ClassDef(
        self, original_node: libcst.ClassDef, updated_node: libcst.ClassDef
    ) -> libcst.ClassDef:
        # Access the statements inside the indented block
        body_statements = list(updated_node.body.body)

        # Track which function names we're removing
        removed_functions = set()

        # First pass: identify functions to remove
        new_body_statements = []
        for stmt in body_statements:
            if isinstance(stmt, libcst.FunctionDef) and self.flip():
                # Track this function name for removal
                removed_functions.add(stmt.name.value)
                # Skip this function (remove it)
                continue
            new_body_statements.append(stmt)

        # Only proceed if we actually removed something
        if not removed_functions:
            return updated_node

        # Create a reference remover to clean up references to removed functions
        reference_remover = FunctionReferenceRemover(removed_functions)

        # Second pass: process the remaining statements to remove references
        clean_statements = []
        for stmt in new_body_statements:
            # The correct way to apply a transformer to a node
            clean_stmt = stmt.visit(reference_remover)
            clean_statements.append(clean_stmt)

        # Create a new indented block with the cleaned statements
        new_body = updated_node.body.with_changes(body=tuple(clean_statements))

        # Return the updated class with the new body
        return updated_node.with_changes(body=new_body)


class FunctionReferenceRemover(libcst.CSTTransformer):
    """Helper transformer to remove references to deleted functions."""

    def __init__(self, removed_functions):
        super().__init__()
        self.removed_functions = removed_functions
        self.in_self_attr = False

    def visit_Attribute(self, node: libcst.Attribute) -> bool:
        # Check if this is a self.method_name pattern
        if (
            isinstance(node.value, libcst.Name)
            and node.value.value == "self"
            and node.attr.value in self.removed_functions
        ):
            self.in_self_attr = True
        return True

    def leave_Attribute(
        self, original_node: libcst.Attribute, updated_node: libcst.Attribute
    ) -> libcst.BaseExpression:
        if (
            isinstance(updated_node.value, libcst.Name)
            and updated_node.value.value == "self"
            and updated_node.attr.value in self.removed_functions
        ):
            # Reset state
            self.in_self_attr = False
        return updated_node

    def leave_Call(
        self, original_node: libcst.Call, updated_node: libcst.Call
    ) -> libcst.BaseExpression:
        # Check if we're calling a removed function through self
        if self.in_self_attr:
            # Reset state
            self.in_self_attr = False
            # Replace with a placeholder that won't cause errors
            return libcst.Name(value="None")
        return updated_node



import libcst

from swesmith.bug_gen.procedural import PythonProceduralModifier
from swesmith.constants import CodeProperty


FLIPPED_OPERATORS = {
    libcst.Add: libcst.Subtract,
    libcst.And: libcst.Or,
    libcst.BitAnd: libcst.BitOr,
    libcst.BitAnd: libcst.BitXor,
    libcst.BitOr: libcst.BitAnd,
    libcst.BitXor: libcst.BitAnd,
    libcst.Divide: libcst.Multiply,
    libcst.Equal: libcst.NotEqual,
    libcst.FloorDivide: libcst.Modulo,
    libcst.GreaterThan: libcst.LessThan,
    libcst.GreaterThanEqual: libcst.LessThanEqual,
    libcst.In: libcst.NotIn,
    libcst.Is: libcst.IsNot,
    libcst.IsNot: libcst.Is,
    libcst.LeftShift: libcst.RightShift,
    libcst.LessThan: libcst.GreaterThan,
    libcst.LessThanEqual: libcst.GreaterThanEqual,
    libcst.Modulo: libcst.FloorDivide,
    libcst.Multiply: libcst.Divide,
    libcst.NotEqual: libcst.Equal,
    libcst.NotIn: libcst.In,
    libcst.Or: libcst.And,
    libcst.Power: libcst.Multiply,
    libcst.RightShift: libcst.LeftShift,
    libcst.Subtract: libcst.Add,
}


class OperationChangeModifier(PythonProceduralModifier):
    explanation: str = "The operations in an expressions are likely incorrect."
    name: str = "func_pm_op_change"
    conditions: list = [
        CodeProperty.IS_FUNCTION,
        CodeProperty.HAS_BINARY_OP,
    ]

    def leave_BinaryOperation(self, original_node, updated_node):
        if self.flip():
            if isinstance(updated_node.operator, (libcst.Add, libcst.Subtract)):
                updated_node = updated_node.with_changes(
                    operator=self.rand.choice([libcst.Add(), libcst.Subtract()])
                )
            elif isinstance(
                updated_node.operator,
                (libcst.Multiply, libcst.Divide, libcst.FloorDivide, libcst.Modulo),
            ):
                updated_node = updated_node.with_changes(
                    operator=self.rand.choice(
                        [
                            libcst.Multiply(),
                            libcst.Divide(),
                            libcst.FloorDivide(),
                            libcst.Modulo(),
                        ]
                    )
                )
            elif isinstance(
                updated_node.operator, (libcst.BitAnd, libcst.BitOr, libcst.BitXor)
            ):
                updated_node = updated_node.with_changes(
                    operator=self.rand.choice(
                        [libcst.BitAnd(), libcst.BitOr(), libcst.BitXor()]
                    )
                )
            elif isinstance(
                updated_node.operator, (libcst.LeftShift, libcst.RightShift)
            ):
                updated_node = updated_node.with_changes(
                    operator=self.rand.choice([libcst.LeftShift(), libcst.RightShift()])
                )
            elif isinstance(updated_node.operator, (libcst.Power, libcst.Multiply)):
                updated_node = updated_node.with_changes(
                    operator=self.rand.choice([libcst.Power(), libcst.Multiply()])
                )
        return updated_node


class OperationFlipOperatorModifier(PythonProceduralModifier):
    explanation: str = "The operators in an expression are likely incorrect."
    name: str = "func_pm_flip_operators"
    conditions: list = [
        CodeProperty.IS_FUNCTION,
        CodeProperty.HAS_BINARY_OP,
    ]

    def _flip_operator(self, updated_node):
        op_type = type(updated_node.operator)
        if op_type in FLIPPED_OPERATORS:
            # Create a new operator of the flipped type
            new_op_class = FLIPPED_OPERATORS[op_type]
            new_op = new_op_class()

            # Return the binary operation with the flipped operator
            return updated_node.with_changes(operator=new_op)
        return updated_node

    def leave_BinaryOperation(self, original_node, updated_node):
        return self._flip_operator(updated_node) if self.flip() else updated_node

    def leave_BooleanOperation(self, original_node, updated_node):
        return self._flip_operator(updated_node) if self.flip() else updated_node


class OperationSwapOperandsModifier(PythonProceduralModifier):
    explanation: str = "The operands in an expression are likely in the wrong order."
    name: str = "func_pm_op_swap"
    conditions: list = [
        CodeProperty.IS_FUNCTION,
        CodeProperty.HAS_BINARY_OP,
    ]

    def leave_BinaryOperation(self, original_node, updated_node):
        if self.flip():
            updated_node = updated_node.with_changes(
                left=updated_node.right, right=updated_node.left
            )
        return updated_node

    def leave_BooleanOperation(self, original_node, updated_node):
        if self.flip():
            updated_node = updated_node.with_changes(
                left=updated_node.right, right=updated_node.left
            )
        return updated_node


class OperationBreakChainsModifier(PythonProceduralModifier):
    explanation: str = (
        "There are expressions or mathemtical operations that are likely incomplete."
    )
    name: str = "func_pm_op_break_chains"
    conditions: list = [
        CodeProperty.IS_FUNCTION,
        CodeProperty.HAS_BINARY_OP,
    ]

    def leave_BinaryOperation(self, original_node, updated_node):
        if self.flip():
            if isinstance(updated_node.left, libcst.BinaryOperation):
                updated_node = updated_node.with_changes(left=updated_node.left.left)
            elif isinstance(updated_node.right, libcst.BinaryOperation):
                updated_node = updated_node.with_changes(right=updated_node.right.right)
        return updated_node


class OperationChangeConstantsModifier(PythonProceduralModifier):
    explanation: str = "The constants in an expression might be incorrect."
    name: str = "func_pm_op_change_const"
    conditions: list = [
        CodeProperty.IS_FUNCTION,
        CodeProperty.HAS_BINARY_OP,
    ]

    def leave_BinaryOperation(self, original_node, updated_node):
        if self.flip():
            if isinstance(updated_node.left, libcst.Integer):
                try:
                    left_value = int(updated_node.left.value)
                except ValueError:
                    left_value = int(updated_node.left.value, 16)
                updated_node = updated_node.with_changes(
                    left=updated_node.left.with_changes(
                        value=str(left_value + self.rand.choice([-1, 1]))
                    )
                )
            if isinstance(updated_node.right, libcst.Integer):
                try:
                    right_value = int(updated_node.right.value)
                except ValueError:
                    right_value = int(updated_node.right.value, 16)
                updated_node = updated_node.with_changes(
                    right=updated_node.right.with_changes(
                        value=str(right_value + self.rand.choice([-1, 1]))
                    )
                )
        return updated_node
